def db_session(fn):
    def func(*args, **kwargs):
        sesmaker = args[0].sessionmaker
        session = sesmaker()
        fn(*args, dbsession=session, **kwargs)
        session.close()
    return func



import falcon
import json
import sqlalchemy
from .decorators import db_session


class ModelCollection:
    def __init__(self, sessionmaker=None):
        self.sessionmaker = sessionmaker

    @db_session
    def on_get(self, req, resp, dbsession):
        params = {
            'limit': None,
            'offset': 0,
            'order_by': None
        }
        params.update(req.params)
        result = dbsession.query(self.model_class)
        filter = self._prepare_filter(params)
        if len(filter) > 0:
            result = result.filter(*filter)
        if params['order_by'] is not None:
            result = result.order_by(params['order_by'])
        if params['limit'] is not None:
            result = result.limit(params['limit'])
        result = result.offset(params['offset'])
        resp.body = result.all()

    @db_session
    def on_post(self, req, resp, dbsession):
        resource = self.model_class()
        data = json.loads(req.bounded_stream.read().decode())
        for k, v in data.items():
            if hasattr(self.model_class.__table__.columns, k):
                setattr(resource, k, v)
        dbsession.add(resource)
        dbsession.commit()
        dbsession.refresh(resource)
        resp.body = resource

    def _prepare_filter(self, params):
        filter = []
        for k, v in params.items():
            column = getattr(self.model_class.__table__.columns, k, None)
            if column is not None:
                filter.append(column == v)
            else:
                split = k.rsplit('__', 1)
                if len(split) != 2:
                    continue
                else:
                    col = split[0]
                    op = split[1]
                column = getattr(self.model_class.__table__.columns, col, None)
                if column is None:
                    continue
                if isinstance(v, str):
                    if op == 'startswith':
                        filter.append(column.startswith(v))
                    elif op == 'endswith':
                        filter.append(column.endswith(v))
                    elif op == 'contains':
                        filter.append(column.contains(v))
                if op == 'lt':
                    filter.append(column < float(v))
                elif op == 'gt':
                    filter.append(column > float(v))
                elif op == 'lte':
                    filter.append(column <= float(v))
                elif op == 'gte':
                    filter.append(column >= float(v))
        return filter


class ModelResource:
    def __init__(self, sessionmaker=None):
        self.sessionmaker = sessionmaker

    @db_session
    def on_get(self, req, resp, dbsession, **kwargs):
        result = dbsession.query(self.model_class)
        result = result.filter_by(**kwargs)
        try:
            resp.body = result.one()
        except sqlalchemy.orm.exc.NoResultFound:
            raise falcon.HTTPNotFound

    @db_session
    def on_patch(self, req, resp, dbsession, **kwargs):
        result = dbsession.query(self.model_class)
        result = result.filter_by(**kwargs)
        try:
            resource = result.one()
        except sqlalchemy.orm.exc.NoResultFound:
            raise falcon.HTTPNotFound
        data = json.loads(req.bounded_stream.read().decode())
        for k, v in data.items():
            if hasattr(self.model_class.__table__.columns, k):
                setattr(resource, k, v)
            else:
                raise falcon.HTTPBadRequest(
                    description=('Resource does not contain field'
                                 '"{}"'.format(k)))
        dbsession.commit()



import pytest
import json
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.schema import Column
from sqlalchemy.types import Text, Integer
from falcon import testing, API
from falcon_hateoas import resources
from unittest import mock


@pytest.fixture
def sessionmaker():
    sm = mock.MagicMock()
    return sm


@pytest.fixture
def resource_mock(sessionmaker):
    res = resources.ModelResource(sessionmaker)
    filter_by = sessionmaker.return_value.query.return_value.filter_by
    one = filter_by.return_value.one
    one.side_effect = [json.dumps({'some': 'data'}), NoResultFound]
    res.model_class = mock.MagicMock()
    res.model_class.__table__ = mock.DEFAULT
    res.model_class.__table__.columns = mock.DEFAULT
    return res


@pytest.fixture
def collection_mock(sessionmaker):
    res = resources.ModelCollection(sessionmaker)
    query = sessionmaker.return_value.query
    offset = query.return_value.offset
    all = offset.return_value.all
    all.return_value = json.dumps({'some': 'data'})

    order_by = query.return_value.order_by
    limit = order_by.return_value.limit
    offset = limit.return_value.offset
    all = offset.return_value.all
    all.return_value = ''

    filter = query.return_value.filter
    offset = filter.return_value.offset
    all = offset.return_value.all
    all.return_value = ''

    res.model_class = mock.MagicMock()
    res.model_class.__table__ = mock.DEFAULT
    res.model_class.__table__.columns = mock.DEFAULT
    return res


@pytest.fixture
def app(resource_mock, collection_mock):
    app = API()
    app.add_route('/{res_id}', resource_mock)
    app.add_route('/', collection_mock)
    return app


class TestModelResource:
    def test_get(self, app):
        resp = testing.simulate_get(app, '/123')
        assert resp.json == {'some': 'data'}
        resp = testing.simulate_get(app, '/0')
        assert resp.status_code == 404
        resp = testing.simulate_post(app, '/123', body='somedata')
        assert resp.status_code == 405

    def test_patch(self, app):
        res = app._router.find('/123')[0]
        res.model_class.__table__.columns.some_str = str
        session = res.sessionmaker()
        res.sessionmaker.return_value.query.return_value.filter_by\
            .return_value.one.side_effect = None
        # res.sessionmaker.return_value.query.return_value.filter_by\
        #     .return_value.one.return_value = 123
        patch_data = {'some_str': 'thing'}
        resp = testing.simulate_patch(app, '/123', body=json.dumps(patch_data))
        assert resp.status_code == 200
        calls = [
            mock.call.query(res.model_class),
            mock.call.query().filter_by(res_id='123'),
            mock.call.query().filter_by().one(),
            mock.call.commit(),
            mock.call.close()
        ]
        assert session.mock_calls == calls


class TestModelCollection:
    def test_basic_test(self, app):
        resp = testing.simulate_get(app, '/')
        res = app._router.find('/')[0]
        session = res.sessionmaker()
        calls = [
            mock.call.query(res.model_class),
            mock.call.query(res.model_class).offset(0),
            mock.call.query(res.model_class).offset(0).all(),
            mock.call.close()
        ]
        assert session.mock_calls == calls
        assert resp.json == {'some': 'data'}

    def test_meta(self, app):
        params = {
            'limit': 10,
            'offset': 5,
            'order_by': 'some_col'
        }
        res = app._router.find('/')[0]
        session = res.sessionmaker()
        testing.simulate_get(app, '/', params=params)
        query = mock.call.query(res.model_class)
        calls = [
            query,
            query.order_by('some_col'),
            query.order_by('some_col').limit('10'),
            query.order_by('some_col').limit('10').offset('5'),
            query.order_by('some_col').limit('10').offset('5').all(),
            mock.call.close()
        ]
        assert session.mock_calls == calls

    def test_filter(self, app):
        params = {
            'some_col': 'some_val',
            'some_col__startswith': 'some',
            'some_col__endswith': 'val',
            'some_col__contains': '_',
            'some_num__lt': 6,
            'some_num__gt': 4,
            'some_num__lte': 5,
            'some_num__gte': 5,
            'fake_col__contains': 'x'
        }
        res = app._router.find('/')[0]
        session = res.sessionmaker()
        res.model_class.__table__.columns.some_col = Column(Text)
        res.model_class.__table__.columns.some_num = Column(Integer)
        testing.simulate_get(app, '/', params=params)
        query = mock.call.query(res.model_class)
        calls = [
            query,
            query.filter(*([mock.ANY])*8),
            query.filter().offset(0),
            query.filter().offset().all(),
            mock.call.close()
        ]
        assert session.mock_calls == calls

    def test_post(self, app):
        res = app._router.find('/')[0]
        res.model_class.__table__.columns.some_str = str
        session = res.sessionmaker()
        post_data = {
            'some_str': 'data',
            'some_int': 123
        }
        m = res.model_class.return_value
        p1 = mock.PropertyMock()
        setattr(type(m), 'some_str', p1)
        p2 = mock.PropertyMock()
        setattr(type(m), 'some_int', p2)
        m.encode.return_value = json.dumps(post_data).encode()
        rsp = testing.simulate_post(app, '/', body=json.dumps(post_data))
        assert rsp.status_code == 200
        assert res.model_class.called
        assert p1.mock_calls == [mock.call('data')]
        assert p2.mock_calls == []
        calls = [
            mock.call.add(res.model_class()),
            mock.call.commit(),
            mock.call.refresh(res.model_class()),
            mock.call.close()
        ]
        assert session.mock_calls == calls
        assert rsp.json == post_data



import json
import decimal
import datetime


class AlchemyJSONEncoder(json.JSONEncoder):
    def default(self, o):
        d = {}
        for col in o.__table__.columns.keys():
            value = getattr(o, col)
            if hasattr(value, 'isoformat'):
                d[col] = value.isoformat()
            elif isinstance(value, datetime.timedelta):
                d[col] = str(value)
            elif isinstance(value, decimal.Decimal):
                d[col] = float(value)
            else:
                d[col] = value
        return d


class JsonMiddleware:
    def process_response(self, req, resp, resource):
        resp.set_header('Content-Type', 'application/json; charset=utf-8')
        resp.body = json.dumps(resp.body, cls=AlchemyJSONEncoder)
